
class ConnectionException(Exception):
    pass


class ContextIsLockedException(Exception):
    pass


class FileAlreadyExistsException(Exception):
    pass


class FileSystemNotSupported(Exception):
    pass
